@@ -4,6 +4,12 @@
 #include "../../ext/sdot/src/sdot/Support/Assert.cpp"
 #include "../../ext/sdot/src/sdot/Support/Mpi.cpp"
 
+#include "../../ext/sdot/src/sdot/Integration/Arfd.cpp"
+
+#ifdef PD_WANT_STAT
+#include "../../ext/sdot/src/sdot/Support/Stat.cpp"
+#endif
+
 #include "../../ext/sdot/src/sdot/Domains/ConvexPolyhedronAssembly.h"
 #include "../../ext/sdot/src/sdot/Domains/ScaledImage.h"
 
@@ -12,13 +18,16 @@
 #include "../../ext/sdot/src/sdot/PowerDiagram/get_der_centroids_and_integrals_wrt_weight_and_positions.h"
 #include "../../ext/sdot/src/sdot/PowerDiagram/get_der_integrals_wrt_weights.h"
 // #include "../../ext/sdot/src/sdot/PowerDiagram/get_der_boundary_integral.h"
+#include "../../ext/sdot/src/sdot/PowerDiagram/get_distances_from_boundaries.h"
 #include "../../ext/sdot/src/sdot/PowerDiagram/get_boundary_integral.h"
+#include "../../ext/sdot/src/sdot/PowerDiagram/get_image_integrals.h"
 #include "../../ext/sdot/src/sdot/PowerDiagram/get_integrals.h"
 #include "../../ext/sdot/src/sdot/PowerDiagram/get_centroids.h"
 
 #include "../../ext/sdot/src/sdot/Display/VtkOutput.h"
 #include "inferno_color_map.h"
 
+#include <pybind11/functional.h>
 #include <pybind11/pybind11.h>
 #include <pybind11/numpy.h>
 #include <pybind11/stl.h>
@@ -31,53 +40,107 @@ namespace {
     template<class FU>
     void find_radial_func( const std::string &func, const FU &fu ) {
         if ( func == "1" || func == "unit" ) {
-            fu( sdot::FunctionEnum::Unit() );
+            sdot::FunctionEnum::Unit f;
+            fu( f );
             return;
         }
 
         if ( func.size() > 13 && func.substr( 0, 13 ) == "exp((w-r**2)/" ) {
             PD_TYPE eps;
             std::istringstream is( func.substr( 13, func.size() - 14 ) );
+
             is >> eps;
-            fu( sdot::FunctionEnum::ExpWmR2db<PD_TYPE>{ eps } );
+            sdot::FunctionEnum::ExpWmR2db<PD_TYPE> f{ eps };
+            fu( f );
             return;
         }
 
         if ( func == "r**2" || func == "r^2" ) {
-            fu( sdot::FunctionEnum::R2() );
+            sdot::FunctionEnum::R2 f;
+            fu( f );
+            return;
+        }
+
+        if ( func == "pos_part(w-r**2)" || func == "pos_part(w-r^2)" ) {
+            sdot::FunctionEnum::PpWmR2 f;
+            fu( f );
             return;
         }
 
         if ( func == "in_ball(weight**0.5)" ) {
-            fu( sdot::FunctionEnum::InBallW05() );
+            sdot::FunctionEnum::InBallW05 f;
+            fu( f );
+            return;
+        }
+
+        if ( func == "r**2*in_ball(weight**0.5)" || func == "r^2*in_ball(weight**0.5)" ) {
+            sdot::FunctionEnum::R2InBallW05 f;
+            fu( f );
             return;
         }
 
         throw pybind11::value_error( "unknown function type" );
     }
 
-    template<class Domain,class Grid>
-    pybind11::array_t<TF> get_integrals( pybind11::array_t<TF> &positions, pybind11::array_t<TF> &weights, Domain &domain, Grid &grid, const std::string &func ) {
-        auto buf_positions = positions.request();
-        auto buf_weights = weights.request();
+    template<class RF,class FU>
+    void find_radial_func( const RF &func, const FU &fu ) {
+        pybind11::gil_scoped_release release{};
+        fu( func );
+    }
 
-        auto ptr_positions = reinterpret_cast<const typename Grid::Pt *>( buf_positions.ptr );
-        auto ptr_weights = reinterpret_cast<const typename Grid::TF *>( buf_weights.ptr );
+    template<class Domain,class Grid,class FUNC>
+    pybind11::array_t<TF> get_integrals( pybind11::array_t<TF> &positions, pybind11::array_t<TF> &weights, Domain &domain, Grid &grid, const FUNC &func ) {
+        auto ptr_positions = reinterpret_cast<const typename Grid::Pt *>( positions.data() );
+        auto ptr_weights = reinterpret_cast<const typename Grid::TF *>( weights.data() );
 
         pybind11::array_t<TF> res;
         res.resize( { positions.shape( 0 ) } );
         auto buf_res = res.request();
         auto ptr_res = (TF *)buf_res.ptr;
 
-        find_radial_func( func, [&]( auto ft ) {
+        find_radial_func( func, [&]( const auto &ft ) {
             sdot::get_integrals( ptr_res, grid, domain, ptr_positions, ptr_weights, positions.shape( 0 ), ft );
         } );
 
         return res;
     }
 
-    template<class Domain,class Grid>
-    pybind11::array_t<TF> get_centroids( pybind11::array_t<TF> &positions, pybind11::array_t<TF> &weights, Domain &domain, Grid &grid, const std::string &func ) {
+    template<class Domain,class Grid,class FUNC>
+    pybind11::array_t<TF> get_image_integrals( pybind11::array_t<TF> &positions, pybind11::array_t<TF> &weights, Domain &domain, Grid &grid, const FUNC &func, pybind11::array_t<TF> &beg, pybind11::array_t<TF> &end, pybind11::array_t<std::size_t> &nbp ) {
+        auto ptr_positions = reinterpret_cast<const typename Grid::Pt *>( positions.data() );
+        auto ptr_weights = reinterpret_cast<const typename Grid::TF *>( weights.data() );
+        auto ptr_nbp = reinterpret_cast<const std::size_t *>( nbp.data() );
+        auto ptr_beg = reinterpret_cast<const TF *>( beg.data() );
+        auto ptr_end = reinterpret_cast<const TF *>( end.data() );
+        using Pt = typename Grid::Pt;
+        using ST = std::size_t;
+
+        Pt a_beg;
+        Pt a_end;
+        std::array<ST,Grid::dim> a_nbp;
+        for( int d = 0; d < Grid::dim; ++d ) {
+            a_beg[ d ] = ptr_beg[ Grid::dim - 1 - d ];
+            a_end[ d ] = ptr_end[ Grid::dim - 1 - d ];
+            a_nbp[ d ] = ptr_nbp[ Grid::dim - 1 - d ];
+        }
+
+        std::array<std::size_t,Grid::dim+1> shape;
+        for( int d = 0; d < Grid::dim; ++d )
+            shape[ d ] = a_nbp[ d ];
+        shape[ Grid::dim ] = Grid::dim + 1;
+
+        pybind11::array_t<TF> res;
+        res.resize( shape );
+
+        find_radial_func( func, [&]( const auto &ft ) {
+            sdot::get_image_integrals( res.mutable_data(), grid, domain, ptr_positions, ptr_weights, positions.shape( 0 ), ft, a_beg, a_end, a_nbp );
+        } );
+
+        return res;
+    }
+
+    template<class Domain,class Grid,class FUNC>
+    pybind11::array_t<TF> get_centroids( pybind11::array_t<TF> &positions, pybind11::array_t<TF> &weights, Domain &domain, Grid &grid, const FUNC &func, TF rand_ratio ) {
         auto buf_positions = positions.request();
         auto buf_weights = weights.request();
 
@@ -89,11 +152,11 @@ namespace {
         auto buf_res = res.request();
         auto ptr_res = (TF *)buf_res.ptr;
 
-        find_radial_func( func, [&]( auto ft ) {
+        find_radial_func( func, [&]( const auto &ft ) {
             sdot::get_centroids( grid, domain, ptr_positions, ptr_weights, positions.shape( 0 ), ft, [&]( auto centroid, auto, auto num ) {
                 for( int d = 0; d < PD_DIM; ++d )
                     ptr_res[ PD_DIM * num + d ] = centroid[ d ];
-            } );
+            }, rand_ratio );
         } );
 
         return res;
@@ -118,7 +181,7 @@ namespace {
     };
 
     struct PyPc {
-        static constexpr int allow_translations = 0;
+        static constexpr int allow_translations = 1;
         static constexpr int nb_bits_per_axis   = 31;
         static constexpr int allow_ball_cut     = 1;
         static constexpr int dim                = PD_DIM;
@@ -293,14 +356,20 @@ namespace {
         using TI = typename TB::TI;
 
         PyScaledImage( pybind11::array_t<PD_TYPE> &min_pt, pybind11::array_t<PD_TYPE> &max_pt, pybind11::array_t<PD_TYPE> &img ) {
-            if ( min_pt.size() != PyPc::dim )
+            if ( min_pt.size() != dim )
                 throw pybind11::value_error( "wrong dimensions for point" );
-            if ( max_pt.size() != PyPc::dim )
+            if ( max_pt.size() != dim )
                 throw pybind11::value_error( "wrong dimensions for point" );
+
             std::array<TI,dim> sizes;
             for( std::size_t d = 0; d < dim; ++d )
-                sizes[ d ] = img.shape( dim - 1 - d );
-            bounds = { Pt( min_pt.data() ), Pt( max_pt.data() ), img.data(), sizes };
+                sizes[ d ] = img.shape( img.ndim() - 1 - d );
+
+            TI nb_coeffs = 1;
+            if ( img.ndim() == dim + 1 )
+                nb_coeffs = img.shape( 0 );
+
+            bounds = { Pt( min_pt.data() ), Pt( max_pt.data() ), img.data(), sizes, nb_coeffs };
         }
 
         // PD_TYPE coeff_at( pybind11::array_t<PD_TYPE> &point ) {
@@ -351,8 +420,8 @@ namespace {
         TB bounds;
     };
 
+    #define PyPowerDiagramZGrid PyPowerDiagramZGrid_##PD_DIM
 
-    // template<int dim,class TF>
     struct PyPowerDiagramZGrid {
         using Grid = sdot::SpZGrid<PyPc>;
         using Pt   = typename Grid::Pt;
@@ -360,40 +429,30 @@ namespace {
         PyPowerDiagramZGrid( int max_dirac_per_cell ) : grid( max_dirac_per_cell ) {
         }
 
-        void update( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, bool positions_have_changed, bool weights_have_changed, std::string radial_func ) {
+        void add_replication( pybind11::array_t<PD_TYPE> &positions ) {
+            Pt p;
+            for( int d = 0; d < PD_DIM; ++d )
+                p[ d ] = positions.at( d );
+            grid.translations.push_back( p );
+        }
+
+        void update( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, bool positions_have_changed, bool weights_have_changed, bool ball_cut ) {
             auto buf_positions = positions.request();
             auto buf_weights = weights.request();
             if ( buf_positions.shape[ 1 ] != PyPc::dim )
                 throw pybind11::value_error( "dim does not correspond to shape[ 1 ] of positions" );
-
             grid.update(
                 reinterpret_cast<const Pt *>( buf_positions.ptr ),
                 reinterpret_cast<const TF *>( buf_weights.ptr ),
                 positions.shape( 0 ),
                 positions_have_changed,
                 weights_have_changed,
-                radial_func == "in_ball(weight**0.5)"
+                ball_cut
             );
         }
 
-        pybind11::array_t<PD_TYPE> integrals_acp( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyConvexPolyhedraAssembly<dim,TF> &domain, const std::string &func ) {
-            return get_integrals( positions, weights, domain.bounds, grid, func );
-        }
-
-        pybind11::array_t<PD_TYPE> integrals_img( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyScaledImage<dim,TF> &domain, const std::string &func ) {
-            return get_integrals( positions, weights, domain.bounds, grid, func );
-        }
-
-        pybind11::array_t<PD_TYPE> centroids_acp( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyConvexPolyhedraAssembly<dim,TF> &domain, const std::string &func ) {
-            return get_centroids( positions, weights, domain.bounds, grid, func );
-        }
-
-        pybind11::array_t<PD_TYPE> centroids_img( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyScaledImage<dim,TF> &domain, const std::string &func ) {
-            return get_centroids( positions, weights, domain.bounds, grid, func );
-        }
-
-        template<class Domain>
-        PyDerResult<dim,TF> der_integrals_wrt_weights( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const std::string &func, bool stop_if_void ) {
+        template<class Domain,class FUNC>
+        PyDerResult<dim,TF> der_integrals_wrt_weights( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const FUNC &func, bool stop_if_void ) {
             auto buf_positions = positions.request();
             auto buf_weights = weights.request();
 
@@ -406,7 +465,7 @@ namespace {
             std::vector<PD_TYPE    > w_v_values;
 
             PyDerResult<dim,TF> res;
-            find_radial_func( func, [&]( auto ft ) {
+            find_radial_func( func, [&]( const auto &ft ) {
                 res.error = sdot::get_der_integrals_wrt_weights( w_m_offsets, w_m_columns, w_m_values, w_v_values, grid, domain.bounds, ptr_positions, ptr_weights, std::size_t( positions.shape( 0 ) ), ft, stop_if_void );
             } );
 
@@ -418,15 +477,24 @@ namespace {
             return res;
         }
 
-        PyDerResult<dim,TF> der_integrals_wrt_weights_acp( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyConvexPolyhedraAssembly<dim,TF> &domain, const std::string &func, bool stop_if_void ) {
-            return der_integrals_wrt_weights( positions, weights, domain, func, stop_if_void );
-        }   
+        template<class Domain,class FUNC>
+        pybind11::array_t<TF> distances_from_boundaries( pybind11::array_t<PD_TYPE> &points, pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const FUNC &func, bool count_domain_boundaries ) {
+            auto ptr_positions = reinterpret_cast<const Pt *>( positions.data() );
+            auto ptr_points = reinterpret_cast<const Pt *>( points.data() );
+            auto ptr_weights = weights.data();
 
-        PyDerResult<dim,TF> der_integrals_wrt_weights_img( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyScaledImage<dim,TF> &domain, const std::string &func, bool stop_if_void ) {
-            return der_integrals_wrt_weights( positions, weights, domain, func, stop_if_void );
-        }   
-    
-        PyDerResult<dim,TF> der_centroids_and_integrals_wrt_weight_and_positions( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyConvexPolyhedraAssembly<dim,TF> &domain, const std::string &func ) {
+            pybind11::array_t<PD_TYPE> res;
+            res.resize( { points.shape( 0 ) } );
+
+            find_radial_func( func, [&]( const auto &ft ) {
+                sdot::get_distances_from_boundaries( res.mutable_data(), ptr_points, std::size_t( points.shape( 0 ) ), grid, domain.bounds, ptr_positions, ptr_weights, std::size_t( positions.shape( 0 ) ), ft, count_domain_boundaries );
+            } );
+
+            return res;
+        }
+
+        template<class Domain,class FUNC>
+        PyDerResult<dim,TF> der_centroids_and_integrals_wrt_weight_and_positions( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const FUNC &func ) {
             auto buf_positions = positions.request();
             auto buf_weights = weights.request();
 
@@ -439,7 +507,7 @@ namespace {
             std::vector<PD_TYPE    > w_v_values;
 
             PyDerResult<dim,TF> res;
-            find_radial_func( func, [&]( auto ft ) {
+            find_radial_func( func, [&]( const auto &ft ) {
                 res.error = sdot::get_der_centroids_and_integrals_wrt_weight_and_positions( w_m_offsets, w_m_columns, w_m_values, w_v_values, grid, domain.bounds, ptr_positions, ptr_weights, std::size_t( positions.shape( 0 ) ), ft );
             } );
 
@@ -451,8 +519,8 @@ namespace {
             return res;
         }
 
-        template<class Domain>
-        void display_vtk( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const std::string &func, const char *filename, bool points, bool centroids ) {
+        template<class Domain,class FUNC>
+        void display_vtk( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const FUNC &func, const char *filename, bool points, bool centroids ) {
             //        sdot::VtkOutput<1,TF> vo({ "num" });
             //        grid.display( vo );
             //        vo.save( filename );
@@ -463,11 +531,11 @@ namespace {
             auto ptr_positions = reinterpret_cast<const Pt *>( positions.data() );
             auto ptr_weights = weights.data();
 
-            find_radial_func( func, [&]( auto ft ) {
+            find_radial_func( func, [&]( const auto &ft ) {
                 grid.for_each_laguerre_cell(
                     [&]( auto &lc, std::size_t num_dirac_0, int ) {
                         domain.bounds.for_each_intersection( lc, [&]( auto &cp, auto space_func ) {
-                            if ( space_func.coeff )
+                            if ( space_func )
                                 cp.display( vtk_output, { ptr_weights[ num_dirac_0 ], TF( num_dirac_0 ), TF( 0 ) } );
                         } );
                     },
@@ -487,7 +555,7 @@ namespace {
 
             if ( centroids ) {
                 std::vector<Pt> c( positions.shape( 0 ) );
-                find_radial_func( func, [&]( auto ft ) {
+                find_radial_func( func, [&]( const auto &ft ) {
                     sdot::get_centroids( grid, domain.bounds, ptr_positions, ptr_weights, positions.shape( 0 ), ft, [&]( auto centroid, auto, auto num ) {
                         c[ num ] = centroid;
                     } );
@@ -500,16 +568,8 @@ namespace {
             vtk_output.save( filename );
         }
 
-        void display_vtk_acp( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyConvexPolyhedraAssembly<dim,TF> &domain, const std::string &func, const char *filename, bool points, bool centroids ) {
-            display_vtk( positions, weights, domain, func, filename, points, centroids );
-        }
-
-        void display_vtk_img( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyScaledImage<dim,TF> &domain, const std::string &func, const char *filename, bool points, bool centroids ) {
-            display_vtk( positions, weights, domain, func, filename, points, centroids );
-        }
-
-        template<class Domain>
-        std::string display_html_canvas( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const std::string &func, int hide_after ) {
+        template<class Domain,class FUNC>
+        std::string display_html_canvas( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const FUNC &func, int hide_after ) {
             std::size_t nd = hide_after >= 0 ? std::min( hide_after, int( positions.shape( 0 ) ) ) : positions.shape( 0 );
             auto ptr_positions = reinterpret_cast<const Pt *>( positions.data() );
             auto ptr_weights = reinterpret_cast<const TF *>( weights.data() );
@@ -526,7 +586,7 @@ namespace {
 
             // centroids
             std::vector<Pt> c( positions.shape( 0 ) );
-            find_radial_func( func, [&]( auto ft ) {
+            find_radial_func( func, [&]( const auto &ft ) {
                 sdot::get_centroids( grid, domain.bounds, ptr_positions, ptr_weights, positions.shape( 0 ), ft, [&]( auto centroid, auto, auto num ) {
                     c[ num ] = centroid;
                 } );
@@ -542,11 +602,11 @@ namespace {
             std::vector<Pt> max_pts( thread_pool.nb_threads(), Pt( - std::numeric_limits<TF>::max() ) );
             std::vector<std::ostringstream> os_int( thread_pool.nb_threads() );
             std::vector<std::ostringstream> os_ext( thread_pool.nb_threads() );
-            find_radial_func( func, [&]( auto ft ) {
+            find_radial_func( func, [&]( const auto &ft ) {
                 grid.for_each_laguerre_cell(
                     [&]( auto &lc, std::size_t num_dirac_0, int num_thread ) {
                         domain.bounds.for_each_intersection( lc, [&]( auto &cp, auto space_func ) {
-                            if ( space_func.coeff ) {
+                            if ( space_func ) {
                                 if ( num_dirac_0 < nd ) {
                                     cp.display_html_canvas( os_int[ num_thread ], ptr_weights[ num_dirac_0 ], 0 );
                                     cp.display_html_canvas( os_ext[ num_thread ], ptr_weights[ num_dirac_0 ], 1 );
@@ -593,14 +653,6 @@ namespace {
             return out.str();
         }
 
-        std::string display_html_canvas_acp( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyConvexPolyhedraAssembly<dim,TF> &domain, const std::string &func, int hide_after ) {
-            return display_html_canvas( positions, weights, domain, func, hide_after );
-        }
-
-        std::string display_html_canvas_img( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyScaledImage<dim,TF> &domain, const std::string &func, int hide_after ) {
-            return display_html_canvas( positions, weights, domain, func, hide_after );
-        }
-
         void display_vtk_points( pybind11::array_t<PD_TYPE> &positions, const char *filename ) {
             sdot::VtkOutput<1> vtk_output( { "num" } );
 
@@ -612,7 +664,8 @@ namespace {
             vtk_output.save( filename );
         }
 
-        void display_asy( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, PyConvexPolyhedraAssembly<dim,TF> &domain, const std::string &radial_func, const char *filename, const char *preamble, pybind11::array_t<PD_TYPE> &values, std::string colormap, double linewidth, double dotwidth, bool avoid_bounds, const char *closing, double min_rf, double max_rf ) {
+        template<class Domain,class Func>
+        void display_asy( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, Domain &domain, const Func &radial_func, const char *filename, const char *preamble, pybind11::array_t<PD_TYPE> &values, std::string colormap, double linewidth, double dotwidth, bool avoid_bounds, const char *closing, double min_rf, double max_rf ) {
             auto ptr_positions = reinterpret_cast<const Pt *>( positions.data() );
             #if PD_DIM==2
             auto ptr_weights = weights.data();
@@ -644,7 +697,7 @@ namespace {
             } else {
                 #if PD_DIM==2
                 std::vector<std::ostringstream> outputs( thread_pool.nb_threads() );
-                find_radial_func( radial_func, [&]( auto ft ) {
+                find_radial_func( radial_func, [&]( const auto &ft ) {
                     grid.for_each_laguerre_cell(
                         [&]( auto &lc, std::size_t num_dirac_0, int num_thread ) {
                             domain.bounds.for_each_intersection( lc, [&]( auto &cp, auto space_func ) {
@@ -695,6 +748,37 @@ namespace {
             f << closing;
         }
 
+        #define DEF_FOR( NAME, DOMAIN, FUNC ) \
+            pybind11::array_t<PD_TYPE> integrals_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, FUNC &func ) { \
+                return get_integrals( positions, weights, domain.bounds, grid, func ); \
+            } \
+            pybind11::array_t<PD_TYPE> image_integrals_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &func, pybind11::array_t<TF> &beg, pybind11::array_t<TF> &end, pybind11::array_t<std::size_t> &nb_pixels ) { \
+                return get_image_integrals( positions, weights, domain.bounds, grid, func, beg, end, nb_pixels ); \
+            } \
+            pybind11::array_t<PD_TYPE> centroids_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &func, TF rand_ratio ) { \
+                return get_centroids( positions, weights, domain.bounds, grid, func, rand_ratio ); \
+            } \
+            PyDerResult<dim,TF> der_integrals_wrt_weights_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &func, bool stop_if_void ) { \
+                return der_integrals_wrt_weights( positions, weights, domain, func, stop_if_void ); \
+            } \
+            pybind11::array_t<TF> distances_from_boundaries_##NAME( pybind11::array_t<PD_TYPE> &points, pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &func, bool count_domain_boundaries ) { \
+                return distances_from_boundaries( points, positions, weights, domain, func, count_domain_boundaries ); \
+            } \
+            void display_vtk_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &func, const char *filename, bool points, bool centroids ) { \
+                display_vtk( positions, weights, domain, func, filename, points, centroids ); \
+            } \
+            std::string display_html_canvas_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &func, int hide_after ) { \
+                return display_html_canvas( positions, weights, domain, func, hide_after ); \
+            } \
+            PyDerResult<dim,TF> der_centroids_and_integrals_wrt_weight_and_positions_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &func ) { \
+                return der_centroids_and_integrals_wrt_weight_and_positions( positions, weights, domain, func ); \
+            } \
+            void display_asy_##NAME( pybind11::array_t<PD_TYPE> &positions, pybind11::array_t<PD_TYPE> &weights, DOMAIN<dim,TF> &domain, const FUNC &radial_func, const char *filename, const char *preamble, pybind11::array_t<PD_TYPE> &values, std::string colormap, double linewidth, double dotwidth, bool avoid_bounds, const char *closing, double min_rf, double max_rf ) { \
+                display_asy( positions, weights, domain, radial_func, filename, preamble, values, colormap, linewidth, dotwidth, avoid_bounds, closing, min_rf, max_rf ); \
+            }
+        #include "comb_types.h"
+        #undef DEF_FOR
+
         Grid grid;
     };
 }
@@ -704,52 +788,55 @@ PYBIND11_MODULE( PD_MODULE_NAME, m ) {
 
     using DerResult = PyDerResult<PD_DIM,PD_TYPE>;
     pybind11::class_<DerResult>( m, "DerResult" )
-        .def_readwrite( "m_offsets"                                 , &DerResult::m_offsets                                                   , "" )
-        .def_readwrite( "m_columns"                                 , &DerResult::m_columns                                                   , "" )
-        .def_readwrite( "m_values"                                  , &DerResult::m_values                                                    , "" )
-        .def_readwrite( "v_values"                                  , &DerResult::v_values                                                    , "" )
-        .def_readwrite( "error"                                     , &DerResult::error                                                       , "" )
+        .def_readwrite( "m_offsets"                                 , &DerResult::m_offsets                                                                  , "" )
+        .def_readwrite( "m_columns"                                 , &DerResult::m_columns                                                                  , "" )
+        .def_readwrite( "m_values"                                  , &DerResult::m_values                                                                   , "" )
+        .def_readwrite( "v_values"                                  , &DerResult::v_values                                                                   , "" )
+        .def_readwrite( "error"                                     , &DerResult::error                                                                      , "" )
     ;
 
     using ConvexPolyhedraAssembly = PyConvexPolyhedraAssembly<PD_DIM,PD_TYPE>;
     pybind11::class_<ConvexPolyhedraAssembly>( m, "ConvexPolyhedraAssembly" )
-        .def( pybind11::init<>()                                                                                                              , "" )
-        .def( "add_convex_polyhedron"                               , &ConvexPolyhedraAssembly::add_convex_polyhedron                         , "" )
-        .def( "add_box"                                             , &ConvexPolyhedraAssembly::add_box                                       , "" )
-        .def( "normalize"                                           , &ConvexPolyhedraAssembly::normalize                                     , "" )
-        .def( "display_boundaries_vtk"                              , &ConvexPolyhedraAssembly::display_boundaries_vtk                        , "" )
-        .def( "min_position"                                        , &ConvexPolyhedraAssembly::min_position                                  , "" )
-        .def( "max_position"                                        , &ConvexPolyhedraAssembly::max_position                                  , "" )
-        .def( "coeff_at"                                            , &ConvexPolyhedraAssembly::coeff_at                                      , "" )
-        .def( "measure"                                             , &ConvexPolyhedraAssembly::measure                                       , "" )
+        .def( pybind11::init<>()                                                                                                                             , "" )
+        .def( "add_convex_polyhedron"                               , &ConvexPolyhedraAssembly::add_convex_polyhedron                                        , "" )
+        .def( "add_box"                                             , &ConvexPolyhedraAssembly::add_box                                                      , "" )
+        .def( "normalize"                                           , &ConvexPolyhedraAssembly::normalize                                                    , "" )
+        .def( "display_boundaries_vtk"                              , &ConvexPolyhedraAssembly::display_boundaries_vtk                                       , "" )
+        .def( "min_position"                                        , &ConvexPolyhedraAssembly::min_position                                                 , "" )
+        .def( "max_position"                                        , &ConvexPolyhedraAssembly::max_position                                                 , "" )
+        .def( "coeff_at"                                            , &ConvexPolyhedraAssembly::coeff_at                                                     , "" )
+        .def( "measure"                                             , &ConvexPolyhedraAssembly::measure                                                      , "" )
     ;
 
     using ScaledImage = PyScaledImage<PD_DIM,PD_TYPE>;
     pybind11::class_<ScaledImage>( m, "ScaledImage" )
-        .def( pybind11::init<pybind11::array_t<PD_TYPE> &, pybind11::array_t<PD_TYPE> &, pybind11::array_t<PD_TYPE> &>()                     , "" )
-        .def( "display_boundaries_vtk"                              , &ScaledImage::display_boundaries_vtk                                   , "" )
-        .def( "min_position"                                        , &ScaledImage::min_position                                             , "" )
-        .def( "max_position"                                        , &ScaledImage::max_position                                             , "" )
-        .def( "coeff_at"                                            , &ScaledImage::coeff_at                                                 , "" )
-        .def( "measure"                                             , &ScaledImage::measure                                                  , "" )
+        .def( pybind11::init<pybind11::array_t<PD_TYPE> &, pybind11::array_t<PD_TYPE> &, pybind11::array_t<PD_TYPE> &>()                                     , "" )
+        .def( "display_boundaries_vtk"                                      , &ScaledImage::display_boundaries_vtk                                           , "" )
+        .def( "min_position"                                                , &ScaledImage::min_position                                                     , "" )
+        .def( "max_position"                                                , &ScaledImage::max_position                                                     , "" )
+        .def( "coeff_at"                                                    , &ScaledImage::coeff_at                                                         , "" )
+        .def( "measure"                                                     , &ScaledImage::measure                                                          , "" )
     ;
 
+
     using PowerDiagramZGrid = PyPowerDiagramZGrid;
     pybind11::class_<PowerDiagramZGrid>( m, "PowerDiagramZGrid" )
-        .def( pybind11::init<int>()                                                                                                           , "" )
-        .def( "update"                                              , &PowerDiagramZGrid::update                                              , "" )
-        .def( "integrals"                                           , &PowerDiagramZGrid::integrals_acp                                       , "" )
-        .def( "integrals"                                           , &PowerDiagramZGrid::integrals_img                                       , "" )
-        .def( "der_integrals_wrt_weights"                           , &PowerDiagramZGrid::der_integrals_wrt_weights_acp                       , "" )
-        .def( "der_integrals_wrt_weights"                           , &PowerDiagramZGrid::der_integrals_wrt_weights_img                       , "" )
-        .def( "der_centroids_and_integrals_wrt_weight_and_positions", &PowerDiagramZGrid::der_centroids_and_integrals_wrt_weight_and_positions, "" )
-        .def( "centroids"                                           , &PowerDiagramZGrid::centroids_acp                                       , "" )
-        .def( "centroids"                                           , &PowerDiagramZGrid::centroids_img                                       , "" )
-        .def( "display_vtk"                                         , &PowerDiagramZGrid::display_vtk_acp                                     , "" )
-        .def( "display_vtk"                                         , &PowerDiagramZGrid::display_vtk_img                                     , "" )
-        .def( "display_html_canvas"                                 , &PowerDiagramZGrid::display_html_canvas_acp                             , "" )
-        .def( "display_html_canvas"                                 , &PowerDiagramZGrid::display_html_canvas_img                             , "" )
-        .def( "display_vtk_points"                                  , &PowerDiagramZGrid::display_vtk_points                                  , "" )
-        .def( "display_asy"                                         , &PowerDiagramZGrid::display_asy                                         , "" )
+        .def( pybind11::init<int>()                                                                                                                          , "" )
+        .def( "update"                                                      , &PowerDiagramZGrid::update                                                     , "" )
+        .def( "add_replication"                                             , &PowerDiagramZGrid::add_replication                                            , "" )
+        #define DEF_FOR( NAME, DOMAIN, FUNC ) \
+                .def( "integrals"                                           , &PowerDiagramZGrid::integrals_##NAME                                           , "" ) \
+                .def( "image_integrals"                                     , &PowerDiagramZGrid::image_integrals_##NAME                                     , "" ) \
+                .def( "der_integrals_wrt_weights"                           , &PowerDiagramZGrid::der_integrals_wrt_weights_##NAME                           , "" ) \
+                .def( "distances_from_boundaries"                           , &PowerDiagramZGrid::distances_from_boundaries_##NAME                           , "" ) \
+                .def( "centroids"                                           , &PowerDiagramZGrid::centroids_##NAME                                           , "" ) \
+                .def( "display_vtk"                                         , &PowerDiagramZGrid::display_vtk_##NAME                                         , "" ) \
+                .def( "display_html_canvas"                                 , &PowerDiagramZGrid::display_html_canvas_##NAME                                 , "" ) \
+                .def( "der_centroids_and_integrals_wrt_weight_and_positions", &PowerDiagramZGrid::der_centroids_and_integrals_wrt_weight_and_positions_##NAME, "" ) \
+                .def( "display_asy"                                         , &PowerDiagramZGrid::display_asy_##NAME                                         , "" )
+        #include "comb_types.h"
+        #undef DEF_FOR
+        .def( "display_vtk_points"                                  , &PowerDiagramZGrid::display_vtk_points                                                 , "" )
     ;
+
 }
