@@ -1,4 +1,6 @@
 from .RadialFuncEntropy import RadialFuncEntropy
+from .RadialFuncPpWmR2 import RadialFuncPpWmR2
 from .RadialFuncInBall import RadialFuncInBall
 from .RadialFuncUnit import RadialFuncUnit
+from .RadialFuncArfd import RadialFuncArfd
 from .RadialFuncR2 import RadialFuncR2
